/*********************************************************************/
/* file: misc.c - misc commands                                      */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992-1996 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 
#include "structs.h"

#include <string.h>
#include <ctype.h>

#include "action.h"
#include "comms_help.h"
#include "llist.h"
#include "main.h"
#include "misc.h"
#include "net.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#include "utility.h"
#include "utils.h"


const uchar *letters = (const uchar *)" 0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]abcdefghijklmnopqrstuvwxyz{|}";
int letters_number = ']' - '0' + '}' - 'a' + 3;

/* externs */

extern struct master_table_data master_table;
extern unsigned char tintin_char;
extern unsigned char tintin_retchar;
extern unsigned char text_start, text_end;
extern unsigned char verbatim_char;
extern int default_chars_per_line;
extern int default_lines_in_scr;
extern unsigned char vars[10][BUFFER_SIZE]; /* the $0, $1, $2,....$9 variables */
extern unsigned char *version_num;
extern struct command_elem *commands;

extern const uchar *CLEARSCREEN;

#ifndef X_DISPLAY_MISSING
#  include <X11/keysym.h>
#endif

/********************/
/* the #all command */
/********************/
session *comm_all(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];
  session *sesptr;
  stack *st, *o_st;

  arg = get_arg_with_spaces(arg, buffer);
  
  o_st = new stack;
  o_st->data = ses;
  st = new stack;
  
  for(sesptr = session::get_sessions(); sesptr; sesptr = (session *)st->data) {
    sesptr->put_nextstack(st);
    if(!(IS_SET(sesptr->ignore_flags, IGNORE_ALL))) {
      parse_input(buffer, sesptr);
    }
  }
  
  delete st;
  if(ses != (const session *)o_st->data) // check if lost orginal session.
    ses = 0;
  delete o_st;
  
  return(ses);
}

// *******************
//  the #bell command
// *******************

session *comm_bell(const uchar *args, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  
  args = get_arg_with_spaces(args, buf);
  if(*buf) {
    if(is_abrev((const uchar *)"on", buf)) {
      master_table.set_flag(MASTER_FLAG_BELL);
      ses->print((const uchar *)"#Bell voice on.\n\r");
      
    } else if(is_abrev((const uchar *)"off", buf)) {
      master_table.rem_flag(MASTER_FLAG_BELL);
      ses->print((const uchar *)"#Bell voice off.\n\r");
      
    } else {
      ses->print(help_bell);
    }
    
  } else {
    ses->print((const uchar *)"\a");
  }
  return(ses);
}

/*********************/
/* the #if   command */
/*********************/
session *comm_if(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], condition[BUFFER_SIZE], thentext[BUFFER_SIZE], elsetext[BUFFER_SIZE];

  arg = get_arg_stop_spaces(arg, condition);
  arg = get_arg_stop_spaces(arg, thentext);
  arg = get_arg_with_spaces(arg, elsetext);
  if(!*thentext) {
    ses->print(help_if);
    return(ses);
  }

  if(string_to_number(condition, ses)) {
    if(IS_SET(ses->echo_flags, ECHO_IF)) {
      sprintf((char *)buf, "[IF: %s]\n\r", thentext);
      ses->print(buf);
    }
    return(parse_input(thentext, ses));
  } else if(*elsetext) {
    if(IS_SET(ses->echo_flags, ECHO_IF)) {
      sprintf((char *)buf, "[IF ELSE: %s]\n\r", elsetext);
      ses->print(buf);
    }
    return(parse_input(elsetext, ses));
  }
  return(ses);
}

// ******************************
//  the #cmp and #while commands
// ******************************

const uchar *parse_compare_flags(int &icase, int &groupvar, int &neg,
				 const uchar *args, session *ses)
{
  uchar buf[BUFFER_SIZE], letter;
  const uchar *tmp, *res_args;
  
  icase = FALSE;
  groupvar = FALSE;
  neg = FALSE;
  
  // only move args when argument is flags
  
  res_args = get_arg_stop_spaces(args, buf);
  if(*buf == '-') {
    args = res_args;
    
    tmp = buf + 1;
    while(isspace(*tmp))
      tmp++;
    
    while(*tmp) {
      switch(*tmp) {
      case 'i':
      case 'I':
	icase = TRUE;
	break;
	
      case 'g':
      case 'G':
	groupvar = TRUE;
	break;
	
      case 'n':
      case 'N':
	neg = TRUE;
	break;

      default:
	letter = *tmp;
	sprintf((char *)buf, "#Unknown flag '-%c' in compare.\n\r",
		letter);
	ses->print(buf);
	break;
      }
      tmp++;
      while(isspace(*tmp))
	tmp++;
    }
    
    res_args = get_arg_stop_spaces(args, buf);
  }
  
  return(args);
}

int parse_compare(const int icase, const int groupvar, const int neg,
		  const uchar *otext1, const uchar *otext2, session *ses)
{
  uchar text1[BUFFER_SIZE], text2[BUFFER_SIZE], buf[BUFFER_SIZE];
  const uchar *tmp;
  int result = FALSE;
  
  substitute_vars(otext1, text1, sizeof(text1), TRUE, ses, TRUE);
  substitute_vars(otext2, text2, sizeof(text2), TRUE, ses, TRUE);

  if(groupvar) {
    tmp = text2;
    if(icase) {
      do {
	tmp = get_arg_stop_spaces(tmp, buf);
	result = (stricmp(text1, buf) == 0);
      } while(*tmp && !result);
    } else {
      do {
	tmp = get_arg_stop_spaces(tmp, buf);
	result = (strcmp((const char *)text1, (const char *)buf) == 0);
      } while(*tmp && !result);
    }
  } else {
    if(icase)
      result = (stricmp(text1, text2) == 0);
    else
      result = (strcmp((const char *)text1, (const char *)text2) == 0);
  }

  if(neg) {
    result = (result ? FALSE : TRUE);
  }
  
  return(result);
}

session *comm_cmp(const unsigned char *arg, session *ses)
{
  uchar text1[BUFFER_SIZE], text2[BUFFER_SIZE];
  unsigned char thentext[BUFFER_SIZE], elsetext[BUFFER_SIZE];
  int icase, groupvar, neg, result;
  
  arg = parse_compare_flags(icase, groupvar, neg, arg, ses);
  arg = get_arg_stop_spaces(arg, text1);
  arg = get_arg_stop_spaces(arg, text2);
  arg = get_arg_stop_spaces(arg, thentext);
  arg = get_arg_with_spaces(arg, elsetext);
  if(!*thentext) {
    ses->print(help_cmp);
    return(ses);
  }
  
  result = parse_compare(icase, groupvar, neg, text1, text2, ses);
  
  if(result) {
    if(IS_SET(ses->echo_flags, ECHO_CMP)) {
      sprintf((char *)text1, "[CMP: %s]\n\r", thentext);
      ses->print(text1);
    }
    return(parse_input(thentext, ses));
  } else if(*elsetext) {
    if(IS_SET(ses->echo_flags, ECHO_CMP)) {
      sprintf((char *)text1, "[CMP ELSE: %s]\n\r", elsetext);
      ses->print(text1);
    }
    return(parse_input(elsetext, ses));
  }
  return(ses);
}

session *comm_while(const uchar *args, session *ses)
{
  uchar text1[BUFFER_SIZE], text2[BUFFER_SIZE], commands[BUFFER_SIZE];
  int icase, groupvar, neg, loops;
  
  args = parse_compare_flags(icase, groupvar, neg, args, ses);
  args = get_arg_stop_spaces(args, text1);
  args = get_arg_stop_spaces(args, text2);
  args = get_arg_with_spaces(args, commands);
  
  if(!*commands) {
    ses->print(help_while);
    return(ses);
  }

  for(loops = 0;
      parse_compare(icase, groupvar, neg, text1, text2, ses) &&
      loops < MAX_WHILE_LOOPS;
      loops++) {
    if(!(ses = parse_input(commands, ses)))
      break;
  }

  if(loops == MAX_WHILE_LOOPS) {
    ses->print((const uchar *)"#ERROR - too many loops in while. (probably forever loop)\n\r");
  }
  
  return(ses);
}

session *comm_for(const unsigned char *arg, session *ses)
{
  unsigned char buf1[BUFFER_SIZE], buf2[BUFFER_SIZE], text[BUFFER_SIZE];
  int num1, num2, i;

  arg = get_arg_stop_spaces(arg, buf1);
  arg = get_arg_stop_spaces(arg, buf2);
  arg = get_arg_with_spaces(arg, text);

  if(!*text) {
    ses->print(help_for);
    return(ses);
  }
  num1 = string_to_number(buf1, ses);
  num2 = string_to_number(buf2, ses);

  for(i = num1; i <= num2; i++) {
    sprintf((char *)vars[0], "%d", i);
    substitute_vars(text, buf1, sizeof(buf1), TRUE, ses, FALSE);
    if(!(ses = parse_input(buf1, ses)))
      break;
  }
  return(ses);
}

/* ******************** * the clear command * ******************** */
session *clear_command(const unsigned char *, session *ses)
{
  ses->get_output()->clear();
  if(ses->get_output() == ses->get_input())
    ses->get_output()->print_prompt(TRUE, FALSE);
  return(ses);
}

/*********************/
/* the #boss command */
/*********************/
session *boss_command(const unsigned char *, session *ses)
{
  char letter;
  int i;
#ifndef X_DISPLAY_MISSING
  basewindow *win;
  
  if(MainDisplay) {
    for(win = basewindow::get_bases(); win; win = win->get_next())
      XIconifyWindow(MainDisplay, XtWindow(win->widget),
		     DefaultScreen(MainDisplay));
  } else {
#endif
    window::max_scroll_region();
    printf("%s", CLEARSCREEN);
    for(i = 0; i < basewindow::get_bases()->rows; i++)
      printf("in-order traverse of tree starting from node %d resulted in %d red nodes\n\r", i, basewindow::get_bases()->rows - i);
    
    /* stop screen from scrolling stuff */
    do {
      letter = getchar();
    } while(letter != '\n' && letter != '\r' && letter != ' ');
    
    redraw_screen();
#ifndef X_DISPLAY_MISSING
  }
#endif
  return(ses);
}

/*********************/
/* the #char command */
/*********************/
session *char_command(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];

  arg = get_arg_stop_spaces(arg, buffer);
  if(ispunct(*buffer)) {
    tintin_char = *buffer;
    ses->print((const uchar *)"#OK. GOT NEW TINTIN-CHAR.\n\r");
  } else
    ses->print((const uchar *)"#SPECIFY A PROPER TINTIN-CHAR! SOMETHING LIKE # OR /!\n\r");

  return(ses);
}

/*********************/
/* the #text command */
/*********************/
session *comm_text(const unsigned char *arg, session *ses)
{
  arg = space_out(arg);
  if(ispunct(*arg) && ispunct(*(arg + 1))) {
    if(*arg != *(arg + 1)) {
      text_start = *arg++;
      text_end =*arg;
      ses->print((const uchar *)"#OK. GOT NEW START AND END CHAR.\n\r");
    } else
      ses->print((const uchar *)"#Start and end char must be diffirent!\n\r");
  } else
    ses->print((const uchar *)"#SPECIFY A PROPER CHARS! SOMETHING LIKE () OR []!\n\r");

  return(ses);
}

/***************************/
/* the #returnchar command */
/***************************/
session *comm_returnchar(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];

  arg = get_arg_stop_spaces(arg, buffer);
  if(ispunct(*buffer)) {
    tintin_retchar = *buffer;
    ses->print((const uchar *)"#OK. GOT NEW TINTIN-RETURN-CHAR.\n\r");
  } else
    ses->print((const uchar *)"#SPECIFY A PROPER TINTIN-RETURN-CHAR! SOMETHING LIKE ; OR :\n\r");
  return(ses);
}

/*********************/
/* the #echo command */
/*********************/
session *echo_command(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], *tmp;
  unsigned char *s_ac = (uchar *)"ACTION";
  unsigned char *s_i  = (uchar *)"IF";
  unsigned char *s_cm = (uchar *)"CMP";
  unsigned char *s_w  = (uchar *)"WAIT";
  unsigned char *s_al = (uchar *)"ALIAS";
  int all = FALSE;
  ulong *echo, flag;

  echo = &ses->echo_flags;

  arg = get_arg_stop_spaces(arg, buf);

  switch(*buf) {
  case 'a': {
    switch(*(buf + 1)) {
    case 'c': {
      flag = ECHO_ACTION;
      tmp = s_ac;
    } break;
    case 'l': {
      if(*(buf + 2) == 'l' && !*(buf + 3)) {
	flag = 0;
	tmp = 0;
	all = TRUE;
      } else {
	flag = ECHO_ALIAS;
	tmp = s_al;
      }
    } break;
    default: {
      ses->print(help_echo);
      return(ses);
    } break;
    } /* switch */
  } break;
  case 'i': {
    flag = ECHO_IF;
    tmp = s_i;
  } break;
  case 'c': {
    flag = ECHO_CMP;
    tmp = s_cm;
  } break;
  case 'w': {
    flag = ECHO_WAIT;
    tmp = s_w;
  } break;
  default: {
    ses->print(help_echo);
    return(ses);
  } break;
  } /* switch */

  arg = get_arg_with_spaces(arg, buf);
  if(*buf || all) {
    if(is_abrev((const uchar *)"on", buf) ||
       is_abrev((const uchar *)"yes", buf)) {
      if(all) {
	SET_BIT(*echo, ECHO_ACTION | ECHO_ALIAS | ECHO_WAIT | ECHO_IF | ECHO_CMP);
	sprintf((char *)buf, "#OK. ALL ECHO IS NOW ON.\n\r");
      } else {
	SET_BIT(*echo, flag);
	sprintf((char *)buf, "#OK. %s ECHO IS NOW ON.\n\r", tmp);
      }
    } else if(is_abrev((const uchar *)"off", buf) ||
	      is_abrev((const uchar *)"no", buf)) {
      if(all) {
	*echo = 0;
	sprintf((char *)buf, "#OK. ALL ECHO IS NOW OFF.\n\r");
      } else {
	REMOVE_BIT(*echo, flag);
	sprintf((char *)buf, "#OK. %s ECHO IS NOW OFF.\n\r", tmp);
      }
    } else {
      sprintf((char *)buf, "#Use 'on', 'yes', 'off' or 'no'\n\r");
    }
  } else {
    if(!IS_SET(*echo, flag)) {
      SET_BIT(*echo, flag);
      sprintf((char *)buf, "#OK. %s ECHO IS NOW ON.\n\r", tmp);
    } else {
      REMOVE_BIT(*echo, flag);
      sprintf((char *)buf, "#OK. %s ECHO IS NOW OFF.\n\r", tmp);
    }
  }
  ses->print(buf);
  return(ses);
}

// ********************
// the #message command
// ********************

struct message_elem {
  const uchar *name;
  ulong bit;
  const uchar *descr;
} msgs[] = {
  {(const uchar *)"action",     MESSAGE_ACTION,
   (const uchar *)"Action define message on/off."},
  {(const uchar *)"alias",      MESSAGE_ALIAS,
   (const uchar *)"Alias define message in/off."},
  {(const uchar *)"macro",      MESSAGE_MACRO,
   (const uchar *)"Macro defined message on/off."},
  {(const uchar *)"path",       MESSAGE_PATH,
   (const uchar *)"Path adding message on/off."},
  {(const uchar *)"set",        MESSAGE_SET,
   (const uchar *)"Set command message on/off."},
  {(const uchar *)"substitute", MESSAGE_SUBS,
   (const uchar *)"Substitute define message on/off."},
  {(const uchar *)"variable",   MESSAGE_VAR,
   (const uchar *)"Variable define message on/off."},
  {(const uchar *)"wait",       MESSAGE_WAIT,
   (const uchar *)"Wait define message on/off."},
  {(const uchar *)"word",       MESSAGE_WORD,
   (const uchar *)"Word adding message on/off."},
  {0, 0, 0}
};

session *comm_message(const unsigned char *arg, session *ses)
{
  uchar buf[BUFFER_SIZE];
  int pos;
  
  arg = get_arg_stop_spaces(arg, buf);
  
  if(!*buf) {
    ses->print(help_message);
    ses->print((const uchar *)"#Where field can be any of following:\n\r");
    for(pos = 0; msgs[pos].name; pos++) {
      sprintf((char *)buf, "# %-10s %-5s  : %s\n\r",
	      msgs[pos].name,
	      (IS_SET(ses->message_flags, msgs[pos].bit) ? "(on)" : "(off)"),
	      msgs[pos].descr);
      ses->print(buf);
    }
    ses->print((const uchar *)"# all               : All message options on/off.\n\r");
    
  } else {
    
    for(pos = 0; msgs[pos].name; pos++) {
      if(is_abrev(msgs[pos].name, buf))
	break;
    }
    
    if(!msgs[pos].name && strcmp("all", (const char *)buf)) {
      ses->print((const uchar *)"# Unknown field type #message without arguments to get allowed fields.\n\r");
      return(ses);
    }
    
    arg = get_arg_with_spaces(arg, buf);
    
    if(*buf || !msgs[pos].name) {
      
      if(is_abrev((const uchar *)"on", buf) ||
	 is_abrev((const uchar *)"yes", buf)) {
	if(!msgs[pos].name) { /* all */
	  ses->message_flags = MESSAGE_ALL;
	  sprintf((char *)buf, "#OK. ALL DEFINE MESSAGES ARE NOW ON.\n\r");
	} else {
	  SET_BIT(ses->message_flags, msgs[pos].bit);
	  sprintf((char *)buf, "#OK. %s DEFINE MESSAGES ARE NOW ON.\n\r",
		  msgs[pos].name);
	}
	
      } else if(is_abrev((const uchar *)"off", buf) ||
		is_abrev((const uchar *)"no", buf)) {
	if(!msgs[pos].name) { /* all */
	  ses->message_flags = 0;
	  sprintf((char *)buf, "#OK. ALL DEFINE MESSAGES ARE NOW OFF.\n\r");
	} else {
	  REMOVE_BIT(ses->message_flags, msgs[pos].bit);
	  sprintf((char *)buf, "#OK. %s DEFINE MESSAGES ARE NOW OFF.\n\r",
		  msgs[pos].name);
	}
	
      } else {
	sprintf((char *)buf, "#Use 'on', 'yes', 'off' or 'no'\n\r");
      }
      
    } else {
      
      if(!IS_SET(ses->message_flags, msgs[pos].bit)) {
	SET_BIT(ses->message_flags, msgs[pos].bit);
	sprintf((char *)buf, "#OK. %s DEFINE MESSAGES ARE NOW ON.\n\r",
		msgs[pos].name);
      } else {
	REMOVE_BIT(ses->message_flags, msgs[pos].bit);
	sprintf((char *)buf, "#OK. %s DEFINE MESSAGES ARE NOW OFF.\n\r",
		msgs[pos].name);
      }
    }
    ses->print(buf);
  }
  
  return(ses);
}

/*********************/
/* the #end command */
/*********************/
session *end_command(const unsigned char *, session *)
{
  exit_tintin(0);
  return(0);
}

/***********************/
/* the #ignore command */
/***********************/
session *comm_ignore(const unsigned char *arg,session *ses)
{
  unsigned char buf[100];
  const uchar *tmp;
  const unsigned char *s_ac = (const uchar *)"ACTION";
  const unsigned char *s_su = (const uchar *)"SUBS";
  const unsigned char *s_al = (const uchar *)"ALIAS";
  const unsigned char *s_all = (const uchar *)"ALL COMMAND";
  const unsigned char *s_va = (const uchar *)"VARIABLE";
  const unsigned char *s_ca = (const uchar *)"CALCULATE";
  ulong *ignore, flag;

  ignore = &ses->ignore_flags;

  arg = space_out(arg);

  switch(*arg) {
  case 'a': {
    switch(*(arg + 1)) {
    case 'c': {
      flag = IGNORE_ACTION;
      tmp = s_ac;
    } break;
    case 'l': {
      if(*(arg + 2) == 'l') {
	flag = IGNORE_ALL;
	tmp = s_all;
      } else {
	flag = IGNORE_ALIAS;
	tmp = s_al;
      }
    } break;
    default: {
      ses->print(help_ignore);
      return(ses);
    } break;
    } /* switch */
  } break;
  case 'c': {
    flag = IGNORE_CALC;
    tmp = s_ca;
  } break;
  case 's': {
    flag = IGNORE_SUBS;
    tmp = s_su;
  } break;
  case 'v': {
    flag = IGNORE_VAR;
    tmp = s_va;
  } break;
  default: {
    ses->print(help_ignore);
    return(ses);
  } break;
  } /* switch */

  if(!IS_SET(*ignore, flag)) {
    SET_BIT(*ignore, flag);
    sprintf((char *)buf, "#OK. %s IGNORE IS NOW ON.\n\r", tmp);
  } else {
    REMOVE_BIT(*ignore, flag);
    sprintf((char *)buf, "#OK. %s IGNORE IS NOW OFF.\n\r", tmp);
  }
  ses->print(buf);

  return(ses);
}

// ********************
//  speedlimit command
// ********************

session *comm_speedlimit(const uchar *arg, session *ses)
{
  uchar buffer[BUFFER_SIZE];
  int newlimit;
  
  arg = get_arg_stop_spaces(arg, buffer);
  if(*buffer) {
    newlimit = string_to_number(buffer, ses);
  } else {
    newlimit = -1;
  }

    
  if(newlimit > 0) {
    sprintf((char *)buffer, "#New speedwalk limit is %d.\n\r", newlimit);
    ses->print(buffer);
    ses->speedlimit = newlimit;
  } else {
    if(ses->speedlimit > 0) {
      ses->print((const uchar *)"#Speedwalk limit off.\n\r");
      ses->speedlimit = -1;
    } else {
      ses->print(help_speedlimit);
    }
  }
  
  return(ses);
}

/***********************/
/* the #system command */
/***********************/
session *system_command(const unsigned char *arg, session *ses)
{
  char letter;
  unsigned char buffer[BUFFER_SIZE];
  
#ifndef X_DISPLAY_MISSING
  if(MainDisplay) {
    ses->print((const uchar *)"#You cannot use system command in XWindow.\n\r");
  } else {
#endif
    sigmask_on();
    
    arg = get_arg_with_spaces(arg, buffer);
    if(*buffer) {
      ses->print((const uchar *)"#OK EXECUTING SHELL COMMAND.\n\r");
      ses->print((const uchar *)"#Press return after it to continue.\n\r");
      
      nocbreak_echo();
      system((const char *)buffer);
      
      do {
	letter = getchar();
      } while(letter != '\n' && letter != '\r');
      
      cbreak_noecho();
      
      ses->print((const uchar *)"#OK COMMAND EXECUTED.\n\r");
      get_screen_size();
    } else {
      
      if(!getenv(DEF_SHELL_ENV)) {
	ses->print((const uchar *)"#ERROR - COULDN'T GET DEFAULT SHELL NAME.\n\r");
      } else {
	
	ses->print((const uchar *)"#OK OPENING DEFAULT SHELL.\n\r");
	nocbreak_echo();
	system(getenv(DEF_SHELL_ENV));
	cbreak_noecho();
	
	ses->print((const uchar *)"#OK SHELL CLOSED.\n\r");
	get_screen_size();
      }
    }
    
    sigmask_off();
#ifndef X_DISPLAY_MISSING
  }
#endif
  return(ses);
}

/********************/
/* the #zap command */
/********************/
session *zap_command(const unsigned char *, session *ses)
{
  window *win;

  win = ses->get_input();
  ses->print((const uchar *)"#ZZZZZZZAAAAAAAAPPPP!!!!!!!!! LET'S GET OUTTA HERE!!!!!!!!\n\r");
  delete ses;
  return(win->get_active_ses());
}

/*********************/
/* the #help command */
/*********************/

session *help_command(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], tmp[1024], buffer[BUFFER_SIZE], *str;
  int pos, comm_no;

  arg = get_arg_stop_spaces(arg, buffer);
  str = buffer;
  if(!*str) {
    sprintf((char *)tmp, "More help by typing: %chelp <COMMAND NAME>\n\r",
	    tintin_char);
    pos = str_cat(buf, 0, BUFFER_SIZE, tmp);
    pos = str_cat(buf, pos, BUFFER_SIZE, (const uchar *)"Possible commands: ");
    for(comm_no = 0; *commands[comm_no].name != '\n'; comm_no++) {
      if(comm_no % 5 == 0)
	pos = str_cat(buf, pos, BUFFER_SIZE, (const uchar *)"\n\r  ");
      sprintf((char *)tmp, "%c%c%-13.13s",
	      commands[comm_no].help_txt ? ' ' : '?',
	      tintin_char,
	      commands[comm_no].name);
      pos = str_cat(buf, pos, BUFFER_SIZE, tmp);
    }
    pos = str_cat(buf, pos, BUFFER_SIZE, (const uchar *)"\n\r");
    ses->print(buf);
  } else {
    do {
      if(*str == tintin_char)
	str++;
      
      for(comm_no = 0; *commands[comm_no].name != '\n'; comm_no++) {
	if(is_abrev(commands[comm_no].name, str)) {
	  if(commands[comm_no].help_txt) {
	    ses->print(commands[comm_no].help_txt);
	  } else {
	    ses->print((const uchar *)"#Sorry no help available for that command.\n\r");
	  }
	  break;
	}
      }
      if(*commands[comm_no].name == '\n') {
	sprintf((char *)tmp, "#There isn't such command as %c%s\n\r", tintin_char, str);
	ses->print(tmp);
      }
      arg = get_arg_stop_spaces(arg, buffer);
      str = buffer;
    } while(*str);
  }
  return(ses);
}

session *comm_resetscreen(const unsigned char *, session *ses)
{
#ifndef X_DISPLAY_MISSING
  if(MainDisplay) {
    ses->print((const uchar *)"#This command won't do anything in XWindow.\n\r");
  } else {
#endif
    cbreak_noecho();
    get_screen_size();
    ses->print((const uchar *)"OK.\n\r");
#ifndef X_DISPLAY_MISSING
  }
#endif
  return(ses);
}

/***********************/
/* the #showme command */
/***********************/
session *comm_showme(const unsigned char *arg, session *ses)
{
  unsigned char tmp[BUFFER_SIZE], result[BUFFER_SIZE], *str;
  int check_act = FALSE;
  unsigned int pos;

  arg = get_arg_stop_spaces(arg, tmp);
  if(*tmp == '-') {
    str = tmp + 1;
    while(isspace(*str))
      str++;
    
    while(*str) {
      switch(*str) {
      case 'c':
      case 'C':
	check_act = TRUE;
	break;
      default:
	sprintf((char *)result, "#Unknown argument (%c).\n\r", *str);
	ses->print(result);
      }
      str++;
      
      while(isspace(*str))
	str++;
    }
    arg = get_arg_with_spaces(arg, tmp);
  }
  
//  pos = substitute_vars(tmp, result, sizeof(result), TRUE, ses, TRUE);
  
  if(check_act) {
    if(!(ses = process_text(tmp, result, pos, ses)))
      return(0);
    
    if(*result || !*tmp)
      pos = str_cat(result, pos, sizeof(result), (const uchar *)"\n\r");
  } else {
    pos = str_cat(result, 0, sizeof(result), tmp);
    pos = str_cat(result, pos, sizeof(result), (const uchar *)"\n\r");
  }
  ses->print(result);
  
  return(ses);
}

/**********************************
 * the #word and #unword commands *
 ******************************** */

session *comm_word(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], tmp[BUFFER_SIZE], **newlist;
  int i, pos;
  
  while(isspace(*arg))
    arg++;
  
  if(*arg) {
    while(*arg) {
      arg = get_arg_stop_spaces(arg, buf);
      
      if(*buf) {
	
	if(ses->word_list_pos >= ses->word_list_size) {
	  ses->word_list_size += 10;
	  
	  newlist = new uchar *[ses->word_list_size];
	  
	  for(i = 0; i < ses->word_list_pos; i++)
	    newlist[i] = ses->word_list[i];
	  
	  if(ses->word_list)
	    delete[] ses->word_list;
	  ses->word_list = newlist;
	}
	
	for(i = 0;
	    i < ses->word_list_pos &&
	    strcmp((const char *)buf, (const char *)ses->word_list[i]) > 0;
	    i++)
	  ;
	
	if(i < ses->word_list_pos &&
	   strcmp((const char *)buf, (const char *)ses->word_list[i]) == 0) {
	  sprintf((char *)tmp, "#Word '%s' is already in list\n\r", buf);
	  ses->print(tmp);
	  
	} else {
	  if(IS_SET(ses->message_flags, MESSAGE_WORD)) {
	    sprintf((char *)tmp, "#Word '%s' added to list.\n\r", buf);
	    ses->print(tmp);
	  }
	  
	  for(pos = i, i = ses->word_list_pos; i > pos; i--)
	    ses->word_list[i] = ses->word_list[i - 1];
	  
	  ses->word_list[pos] = strdup(buf);
	  ses->word_list_pos++;
	}
      }
      while(isspace(*arg))
	arg++;
    }
    
  } else  {
    ses->print(help_word);
  }
  
  return(ses);
}

session *comm_unword(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], tmp[BUFFER_SIZE];
  int pos, found, i;
  
  while(isspace(*arg))
    arg++;
  
  if(*arg) {
    while(*arg) {
      arg = get_arg_stop_spaces(arg, buf);
      
      pos = 0;
      found = FALSE;
      while(pos < ses->word_list_pos) {
	if(mystrcmp(ses->word_list[pos], buf)) {
	  if(IS_SET(ses->message_flags, MESSAGE_WORD)) {
	    sprintf((char *)tmp,
		    "#Word '%s' removed from list.\n\r", ses->word_list[pos]);
	    ses->print(tmp);
	  }
	  free(ses->word_list[pos]);
	  
	  for(i = pos; i < ses->word_list_pos - 1; i++)
	    ses->word_list[i] = ses->word_list[i + 1];
	  ses->word_list[i] = 0;
	  ses->word_list_pos--;
	  
	  found = TRUE;
	} else
	  pos++;
      }
      if(!found) {
	sprintf((char *)tmp, "#No '%s' word on list\n\r", buf);
	ses->print(tmp);
      }
      while(isspace(*arg))
	arg++;
    }
    
  } else {
    ses->print(help_unword);
  }
  
  return(ses);
}

/************************
 * the #version command *
 ********************** */

session *comm_version(const unsigned char *, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  
  sprintf((char *)buf,
	  "********************************************************************\n\r"
	  " ht (HUT TinTin) Version %s\n\r"
	  "\n\r"
	  " If you found bugs or have new ideas what to add write:\n\r"
	  " Email: otulonen@snakemail.hut.fi\n\r"
	  "********************************************************************\n\r",
	  version_num);
  ses->print(buf);
  return(ses);
}

/************************************
 * the #crypt and #decrypt commands *
 ********************************** */

session *comm_crypt(const unsigned char *arg, session *ses)
{
  unsigned char command[BUFFER_SIZE + 20], text[BUFFER_SIZE];
  unsigned char cryptword[BUFFER_SIZE], word[BUFFER_SIZE], letter, *crword = 0;
  int i, j, pos;
  const listnode *ln;
  
  if(ses->get_socket() >= 0) {
    arg = get_arg_stop_spaces(arg, command);
    arg = get_arg_stop_spaces(arg, text);
    arg = get_arg_with_spaces(arg, cryptword);
    
    if(!*text) {
      ses->print((const uchar *)"#No text to crypt.\n\r");
      return(ses);
    }
    
    if(*cryptword) {
      crword = cryptword;
    } else {
      if(!(ln = searchnode_list(ses->variables, (unsigned char *)"HUTCRYPT"))||
	 !*ln->right) {
	ses->print((const uchar *)"#You need to define 'HUTCRYPT' variable first.\n\r");
	return(ses);
      }
      crword = ln->right;
    }
    for(i = 0, j = 0; *(crword + i); i++) {
      letter = *(crword + i);
      if(letter == ' ')
	word[j++] = letter - ' ';
      else if(letter >= '0' && letter <= ']')
	word[j++] = letter - '0' + 1;
      else if(letter >= 'a' && letter <= '}')
	word[j++] = letter - 'a' + ']' - '0' + 2;
    }
    word[j] = '\0';
    
    if(j > 0) {
      pos = str_cat(command, strlen((const char *)command), sizeof(command), (const uchar *)" ~~^");
      
      for(i = 0, j = 0; text[i] != '\0'; i++) {
	letter = text[i];
	if(letter == ' ')
	  letter = letter - ' ';
	else if(letter >= '0' && letter <= ']')
	  letter = letter - '0' + 1;
	else if(letter >= 'a' && letter <= '}')
	  letter = letter - 'a' + ']' - '0' + 2;
	else {
	  if(letter != '~')
	    command[pos++] = letter;
	  continue;
	}
	command[pos++] = letters[(letter + word[j++]) % letters_number];
	if(word[j] == '\0')
	  j = 0;
      }
      pos = str_cat(command, pos, sizeof(command), (const uchar *)"^~~[");
      pos = str_cat(command, pos, sizeof(command), version_num);
      pos = str_cat(command, pos, sizeof(command), (const uchar *)"]");
      ses = ses->write_net(command);
    } else {
      ses->print((const uchar *)"#You have bad 'HUTCRYPT' word define it again\n\r");
    }
  } else {
    ses->print((const uchar *)"#There isn't active session.\n\r");
  }
  return(ses);
}

session *comm_decrypt(const unsigned char *arg, session *ses)
{
  unsigned char text[BUFFER_SIZE], word[BUFFER_SIZE], letter;
  unsigned char cryptword[BUFFER_SIZE], *crword = 0;
  int i, j;
  const listnode *ln;
  
  arg = get_arg_stop_spaces(arg, text);
  arg = get_arg_with_spaces(arg, cryptword);
  
  if(!*text) {
    ses->print((const uchar *)"#No text to decrypt.\n\r");
    return(ses);
  }
  
  if(*cryptword) {
    crword = cryptword;
  } else {
    if(!(ln = searchnode_list(ses->variables, (unsigned char *)"HUTCRYPT")) ||
       !*ln->right) {
      ses->print((const uchar *)"#You need to define 'HUTCRYPT' variable first.\n\r");
      return(ses);
    }
    crword = ln->right;
  }
  for(i = 0, j = 0; *(crword + i); i++) {
    letter = *(crword + i);
    if(letter == ' ')
      word[j++] = letter - ' ';
    else if(letter >= '0' && letter <= ']')
      word[j++] = letter - '0' + 1;
    else if(letter >= 'a' && letter <= '}')
      word[j++] = letter - 'a' + ']' - '0' + 2;
  }
  word[j] = '\0';
  
  if(j > 0) {
    i = 0;
    j = 0;
    while(*(text + i)) {
      if(*(text + i) == '\\') {
	i++;
	if(*(text + i))
	  *(text + j++) = *(text + i++);
      } else
	*(text + j++) = *(text + i++);
    }
    *(text + j) = '\0';
    
    for(i = 0, j = 0; *(text + i); i++) {
      letter = *(text + i);
      
      if(letter == ' ')
	letter = letter - ' ';
      else if(letter >= '0' && letter <= ']')
	letter = letter - '0' + 1;
      else if(letter >= 'a' && letter <= '}')
	letter = letter - 'a' + ']' - '0' + 2;
      else {
	continue;
      }	
      if(letter - word[j] < 0) {
	*(text + i) = letters[letter - word[j] + letters_number];
      } else {
	*(text + i) = letters[letter - word[j]];
      }
      if(word[++j] == '\0')
	j = 0;
    }
    *(text + i) = '\0';
    sprintf((char *)word, "#Decrypted '%s'\n\r", text);
    ses->print(word);
  } else {
    ses->print((const uchar *)"#You have bad 'HUTCRYPT' word, use diffirent one\n\r");
  }
  return(ses);
}

// ****************************
// verbatim char change command
// ****************************

session *comm_verbatim(const unsigned char *arg, session *ses)
{
  unsigned char buffer[BUFFER_SIZE];
  
  arg = get_arg_stop_spaces(arg, buffer);
  if(ispunct(*buffer)) {
    verbatim_char = *buffer;
    ses->print((const uchar *)"#OK. GOT NEW VERBATIM-CHAR.\n\r");
  } else
    ses->print((const uchar *)"#SPECIFY A PROPER VERBATIM-CHAR! SOMETHING LIKE / OR \\!\n\r");
  
  return(ses);
}

// ************************
//  the alive command code
// ************************

session *comm_alive(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  const session *sesptr;
  
  arg = get_arg_with_spaces(arg, buf);
  if(*buf) {
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      if(strcmp((const char *)sesptr->get_name(), (const char *)buf) == 0)
	break;
    
    if(sesptr)
      ses->set_status(TRUE);
    else
      ses->set_status(FALSE);
  } else {
    ses->print(help_alive);
  }
  return(ses);
}

// *****************************
//  the isconected command code
// *****************************

session *comm_isconnected(const uchar *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE];
  const session *sesptr;
  
  arg = get_arg_with_spaces(arg, buf);
  if(*buf) {
    for(sesptr = master_table.main_session;
	sesptr;
	sesptr = sesptr->get_next())
      if(strcmp((const char *)sesptr->get_name(), (const char *)buf) == 0)
	break;
    
    if(sesptr && sesptr->get_socket() >= 0)
      ses->set_status(TRUE);
    else
      ses->set_status(FALSE);
  } else {
    ses->print(help_isconnected);
  }
  return(ses);
}

// ****************
//  random command
// ****************

#ifndef HAVE_RANDOM
#define RANDOM rand()
#else
#define RANDOM random()
#endif

session *comm_random(const unsigned char *arg, session *ses)
{
  unsigned char buf[BUFFER_SIZE], *buf_add[BUFFER_SIZE/2 + 1], *str;
  int pos, num = 0, min;
  
  pos = 0;
  str = buf;
  arg = get_arg_stop_spaces(arg, str);
  while(*str) {
    buf_add[pos++] = str;
    
    while(*str != '\0')
      str++;
    str++;
    
    arg = get_arg_stop_spaces(arg, str);
  }
  if(pos > 0) {
    if(pos > 1) {
#ifdef RAND_MAX
      min = (RAND_MAX/pos) * pos;
      while((num = RANDOM) >= min)
	;
#else
      num = RANDOM;
#endif
      num %= pos;
    }
    
    ses = parse_input(buf_add[num], ses);
  } else {
    ses->print(help_random);
  }
  return(ses);
}

// ***************
//  macro command
// ***************

#ifndef X_DISPLAY_MISSING

#define MODI_NONE  0
#define MODI_SHIFT 1
#define MODI_CTRL  2
#define MODI_ALT   3

struct macro_key_elem {
  const uchar *name;
  KeySym key;
  int type;
} macro_keys[] = {
  {(const uchar *)"F1",    XK_F1,  MODI_NONE},
  {(const uchar *)"F2",    XK_F2,  MODI_NONE},
  {(const uchar *)"F3",    XK_F3,  MODI_NONE},
  {(const uchar *)"F4",    XK_F4,  MODI_NONE},
  {(const uchar *)"F5",    XK_F5,  MODI_NONE},
  {(const uchar *)"F6",    XK_F6,  MODI_NONE},
  {(const uchar *)"F7",    XK_F7,  MODI_NONE},
  {(const uchar *)"F8",    XK_F8,  MODI_NONE},
  {(const uchar *)"F9",    XK_F9,  MODI_NONE},
  {(const uchar *)"F10",   XK_F10, MODI_NONE},
  {(const uchar *)"F11",   XK_F11, MODI_NONE},
  {(const uchar *)"F12",   XK_F12, MODI_NONE},
  {(const uchar *)"shift", 0,      MODI_SHIFT},
  {(const uchar *)"ctrl",  0,      MODI_CTRL},
  {(const uchar *)"alt",   0,      MODI_ALT},
  {0, 0, 0}
};

#endif

session *comm_macro(const unsigned char *arg, session *ses)
{
#ifndef X_DISPLAY_MISSING
  uchar key[BUFFER_SIZE], string[BUFFER_SIZE], *str, *orgstr;
  int i, keynum, shift_key, ctrl_key, alt_key;
  KeySym modifiers[6];
  
  if(!MainDisplay) {
#endif
    ses->print((const uchar *)"#This command isn't in use without X-Window.\n\r");
    return(ses);
#ifndef X_DISPLAY_MISSING
  }
  
  arg = get_arg_stop_spaces(arg, key);
  arg = get_arg_stop_spaces(arg, string);

  keynum    = -1;
  shift_key = FALSE;
  ctrl_key  = FALSE;
  alt_key = FALSE;
  
  str = key;
  
  while(*str) {
    orgstr = str;
    
    while(*str && *str != ' ' && *str != '+' && *str != '-')
      str++;
    
    if(*str) {
      *str++ = '\0';

      while(*str == ' ' || *str == '+' || *str == '-')
	str++;
    }
    
    for(i = 0; macro_keys[i].name; i++) {
      if(!stricmp(macro_keys[i].name, orgstr)) {
	
	switch(macro_keys[i].type) {
	case MODI_NONE:
	  keynum = i;
	  break;
	  
	case MODI_SHIFT:
	  shift_key = TRUE;
	  break;
	  
	case MODI_CTRL:
	  ctrl_key = TRUE;
	  break;
	  
	case MODI_ALT:
	  alt_key = TRUE;
	  break;
	  
	default:
	  abort();
	  break;
	} /* switch */
	
	break;
      }
    }

    if(!macro_keys[i].name) {
      sprintf((char *)string, "# Unknown key or modifier '%s'.\n\r",
	      orgstr);
      ses->print(string);
      return(ses);
    }
  }
	  
  if(keynum < 0 || !*string) {
    ses->print(help_macro);
    return(ses);
  }
  
  for(str = key, orgstr = string; *orgstr; orgstr++) {
    if(*orgstr == '\\') {
      if(*(orgstr + 1)) {
	orgstr++;
	
	switch(*orgstr) {
	case '\\':
	  *str++ = '\\';
	  break;

	case '[':
	  *str++ = '[';
	  break;
	  
	case 'E':
	  *str++ = '\033'; // ESC
	  break;
	  
	case 'n':
	  *str++ = '\n';
	  break;
	  
	case 'r':
	  *str++ = '\r';
	  break;
	  
	default:
	  sprintf((char *)key, "#Uknown \\ secuence (\\%c)\n\r",
		  *orgstr);
	  ses->print(key);
	  return(ses);
	  break;
	  
	} // switch
      }
    } else {
      *str++ = *orgstr;
    }
  }
  *str = '\0';

  if(shift_key || ctrl_key || alt_key) {

    i = 0;
    if(shift_key) {
      modifiers[i++] = XK_Shift_L;
      modifiers[i++] = XK_Shift_R;
    }
    
    if(ctrl_key) {
      modifiers[i++] = XK_Control_L;
      modifiers[i++] = XK_Control_R;
    }

    if(alt_key) {
      modifiers[i++] = XK_Alt_L;
      modifiers[i++] = XK_Alt_R;
    }
    
    XRebindKeysym(MainDisplay, macro_keys[keynum].key, modifiers, i,
		  key, str - key);

  } else {
    XRebindKeysym(MainDisplay, macro_keys[keynum].key, 0, 0, key, str - key);
  }
  
  if(IS_SET(ses->message_flags, MESSAGE_MACRO)) {
    ses->print((const uchar *)"#OK. MACRO DEFINED.\n\r");
  }
#endif

  return(ses);
}

// *************
//  command set
// *************

struct set_args_elem set_args[] = {
  {(const uchar *)"bell",         MASTER_FLAG_BELL,        FALSE,
   "Belling."},
#if defined(HAVE_X11_XAW_DIALOG_H) && defined(HAVE_LIBXAW)
  {(const uchar *)"confirm",      SES_FLAG_ASK_CONFIRM,    TRUE,
   "Confirm before quiting in quit function."},
#endif
  {(const uchar *)"ga_prompt",    SES_FLAG_GA_PROMPT,      TRUE,
   "Wait for new text until line ends with GA or return."},
  {(const uchar *)"emailcheck",   SES_FLAG_EMAILCHECK,     TRUE,
   "Email checking."},
  {(const uchar *)"keybell",      SES_FLAG_KEYBELL,        TRUE,
   "Key bell."},
  {(const uchar *)"linemode",     SES_FLAG_LINEMODE,       TRUE,
   "Linemode."},
  {(const uchar *)"no_neg_off",   MASTER_FLAG_NO_NEG_OFF,  FALSE,
   "Don't use vt100/ansi negative off code."},
  {(const uchar *)"snoop",        SES_FLAG_SNOOP,          TRUE,
   "Snooping on/off."},
  {(const uchar *)"speedwalk",    SES_FLAG_SPEEDWALK,      TRUE,
   "Speedwalking."},
  {(const uchar *)"termecho",     SES_FLAG_TERMECHO,       TRUE,
   "Terminal echoing."},
  {(const uchar *)"ticker",       SES_FLAG_TICKER,         TRUE,
   "Ticker on/off."},
  {(const uchar *)"wfespace",     SES_FLAG_WFESPACE,       TRUE,
   "Add space after word if full word in wordfilling."},
  {(const uchar *)"verbatimmode", SES_FLAG_VERBATIMMODE,   TRUE,
   "Verbatim mode."},
  {(const uchar *)"visual_bell",  MASTER_FLAG_VISUAL_BELL, FALSE,
   "Visual belling."},
  {0, 0, 0}
};

session *comm_set(const unsigned char *args, session *ses)
{
  uchar type[BUFFER_SIZE], value[BUFFER_SIZE];
  int pos;
  struct set_args_elem *set_arg;
  
  args = get_arg_stop_spaces(args, type);
  
  if(!*type) {
    
    ses->print((const uchar *)"# Local:\n\r");
    
    for(pos = 0; set_args[pos].name; pos++) {
      if(set_args[pos].is_local) {
	sprintf((char *)type, "# %-13s %-5s : %s\n\r",
		set_args[pos].name,
		(ses->is_set_flag(set_args[pos].bit) ? "(on)" : "(off)"),
		set_args[pos].descr);
	ses->print(type);
      }
    }
    
    ses->print((const uchar *)"#\n\r# Global:\n\r");
    
    for(pos = 0; set_args[pos].name; pos++) {
      if(!set_args[pos].is_local) {
	sprintf((char *)type, "# %-13s %-5s : %s\n\r",
		set_args[pos].name,
		(master_table.is_set_flag(set_args[pos].bit) ?
		 "(on)" : "(off)"),
		set_args[pos].descr);
	ses->print(type);
      }
    }
    
    return(ses);
  }
  
  set_arg = 0;
  
  for(pos = 0; set_args[pos].name; pos++) {
    if(is_abrev(set_args[pos].name, type)) {
      set_arg = &set_args[pos];
      break;
    }
  }
  
  if(!set_arg) {
    ses->print((const uchar *)"#Unknow field.\n\r");
    ses->print(help_set);
    return(ses);
  }
  
  args = get_arg_stop_spaces(args, value);
  
  if(!*value) {

    if((set_arg->is_local && ses->is_set_flag(set_arg->bit)) ||
       (!set_arg->is_local && master_table.is_set_flag(set_arg->bit))) {
      
      if(IS_SET(ses->message_flags, MESSAGE_SET)) {
	sprintf((char *)value, "# %s set off.\n\r", set_arg->name);
	ses->print(value);
      }
      if(set_arg->is_local)
	ses->rem_flag(set_arg->bit);
      else
	master_table.rem_flag(set_arg->bit);
      
    } else {
      
      if(IS_SET(ses->message_flags, MESSAGE_SET)) {
	sprintf((char *)value, "# %s set on.\n\r", set_arg->name);
	ses->print(value);
      }
      if(set_arg->is_local)
	ses->set_flag(set_arg->bit);
      else
	master_table.set_flag(set_arg->bit);
    }
    
  } else {
    
    if(is_abrev((const uchar *)"on", value)) {
      if(IS_SET(ses->message_flags, MESSAGE_SET)) {
	sprintf((char *)value, "# %s set on.\n\r", set_arg->name);
	ses->print(value);
      }
      
      if(set_arg->is_local)
	ses->set_flag(set_arg->bit);
      else
	master_table.set_flag(set_arg->bit);
      
    } else if(is_abrev((const uchar *)"off", value)) {
      if(IS_SET(ses->message_flags, MESSAGE_SET)) {
	sprintf((char *)value, "# %s set off.\n\r", set_arg->name);
	ses->print(value);
      }
      
      if(set_arg->is_local)
	ses->rem_flag(set_arg->bit);
      else
	master_table.rem_flag(set_arg->bit);
      
    } else {
      
      ses->print(help_set);
    }
  }
  
  return(ses);
}
