#include "color.h"

void redraw_screen(void);
void prompt(void);
void get_screen_size(int &ScrCols, int &ScrRows);
void get_screen_size(void);

void cbreak_noecho(void);
void nocbreak_echo(void);
void init_screen(const int);
void init_termcap();

session *comm_split(const unsigned char *arg, session *ses);
session *comm_open(const unsigned char *arg, session *ses);
session *comm_output(const unsigned char *arg, session *ses);
session *comm_input(const unsigned char *arg, session *ses);
session *comm_close(const unsigned char *arg, session *ses);
session *comm_wininfo(const unsigned char *arg, session *ses);
session *comm_winsize(const unsigned char *arg, session *ses);

window *find_window(const unsigned char *name);

class position {
public:
  position();
  
  const position &operator=(const position &pos);

  int row, col;
};

inline const position &position::operator=(const position &pos)
{
  row = pos.row;
  col = pos.col;
  return(pos);
}

class winhist_elem {
public:
  winhist_elem();
  
  uchar letter;
  unsigned short colormode;
};

class color_clip_reg {
public:
  int s_col, e_col;
  int s_row, e_row;
};

class basewindow {
public:
  basewindow(const uchar *, const char *, const char *, session *);
  ~basewindow();

#ifndef X_DISPLAY_MISSING
  int resize(const int, const int);
#endif
  void check_win_sizes();
  
  static basewindow *get_bases() { return basewindow::basewins; }
  
  basewindow *get_next() const   { return next; }
  
  window *get_windows() const    { return windows; }

  int rows, cols;
  unsigned short colormode;
  struct scrollsize Scroll;
  position Scrcursor;
  
  window *windows;
  
#ifndef X_DISPLAY_MISSING
  Widget widget;
  GC gc;
  XFontStruct *font;
  
  winhist_elem **data;
  int h_row, h_col;
  color_clip_reg *clip_color;
  int x_wait_flush;
#endif
  
  static basewindow *basewins;
  basewindow *next;
};

class timeline_elem {
public:
  timeline_elem();
  ~timeline_elem();

  void set(const uchar *);
  void setlast(const uchar *);
  void setlastcol(const uchar *);
  const uchar *get_defline() const  { return(def_line); }
  const uchar *get_lastline() const { return(last_line); }
  const uchar *get_lastcol() const  { return(last_col); }
  
private:
  uchar *def_line;
  uchar *last_line;
  uchar *last_col;
};

class window {
public:
  window(const uchar *, basewindow *);
  window(const unsigned char *nam, const int size, window *win);
  ~window();

  void move(const int row, const int col, const int flush);
  void move(const position *pos, const int flush) {
    move(pos->row, pos->col, flush);
  }
  void moveforward();
  void movebackward();
  void scroll_region(const int start, const int end);
  void scroll_region(const scrollsize *screg) {
    scroll_region(screg->start, screg->end);
  }
  static void max_scroll_region();
  
  void clear();
  void clearend();
  void clearstart();
  void clearline();
  void clearstartline();
  void clearendline();
  void redraw(const int log = FALSE, const int position = 0);
  // cursor 0 = center, -1 up and 1 = down
  void bell();
  
  void print(const unsigned char *, const unsigned char *add = 0);
  void col_print(const uchar *);
  
  int change_time(const int, const uchar *);
  void print_time(const int all, const int bell = FALSE);
  void print_win_hist_line(const int all, unsigned char *text = 0);
  
  void print_prompt(const int all, const int log = TRUE);
  void remove_user_text();
  
  const int get_hist_pos() const      { return window_hist_pos; }
  void add_hist_pos()                 { window_hist_pos++; }
  void sub_hist_pos()                 { window_hist_pos--; }
  void set_hist_pos(const int n)      { window_hist_pos = n; }
  const session *get_hist_on() const  { return window_hist_on; }
  const int get_hist_clip() const     { return window_hist_clip; }

  void hist_movebackward();
  void hist_moveforward();
  void hist_backward(int amount = 1);
  void hist_forward(int amount = 1);
  void hist_home();
  void hist_end();
  void hist_pageup();
  void hist_pagedown();
  int hist_search(const unsigned char *str, const int forward = 1);
  void hist_set_mark();
  void hist_copy_clip();
  void hist_rem_clip_color();
  int start_win_history(session *ses);
  void exit_win_history();
  void max_size();
  int change_size(const int newsize);
  void next_window_active(window *newact = 0, int focusin = TRUE);
  
  const basewindow *get_base() const      { return base; }
  
  const unsigned char *get_name() const   { return name; }
  
  const scrollsize *get_size() const      { return &WinSize; }
  const scrollsize *get_scroll() const    { return &Scroll; }
  const position *get_cursor() const      { return &cursor; }
#ifndef X_DISPLAY_MISSING
  void set_cursor(int col, int row) { cursor.col = col; cursor.row = row; }
#endif
  const int get_ptexrem() const           { return ptexrem; }
  void set_ptexrem()                      { ptexrem = TRUE; }
  void rem_ptexrem()                      { ptexrem = FALSE; }
  
  void set_active_ses(session *ses);
  session *get_active_ses() const         { return active_ses; }
  
  static window *get_active_win()         { return active; }

  const int get_cols() const                    { return base->cols; }
  const int get_rows() const                    { return base->rows; }
  const unsigned short get_scrcolormode() const { return base->colormode; }
  const position *get_scrcursor() const         { return &base->Scrcursor; }
  
  window *get_next() const        { return next; }

  time_t lastupdate;
  const int get_delay() const     { return timedelay; }
  const int get_time_on() const   { return(timedelay >= 0 ? TRUE : FALSE); }

  const unsigned short get_colormode() const { return colormode; }
  const unsigned short get_savecolor() const { return save_colormode; }
  void set_savecolor(unsigned short newmode) { save_colormode = newmode; }
  void check_color();
  void check_session(const session *ses = 0);
  
#ifndef X_DISPLAY_MISSING
  void xprint(const uchar *, int flush = FALSE);
  void xprint_cursor(int all = 0); /* -1 = remove, 1 = allways */
  void xhist_scrollup();
  void xhist_scrolldown();
#endif

private:
  basewindow *base;

  unsigned char *name;
  unsigned short colormode, save_colormode;
  
  scrollsize WinSize;
  scrollsize Scroll;
  position cursor;
  int timedelay;            // -1 = not active, 0 = no clock
  int ptexrem, autowarp;
  int window_hist_pos, window_hist_clip;
  session *window_hist_on;
  session *active_ses;
  
  timeline_elem timeline;
  
  window *next;

  static window *active;
};

inline void window::check_color()
{
  unsigned char color[BUFFER_SIZE];
  
  bits_to_color(base->colormode, colormode, color);
  if(*color) {
#ifndef X_DISPLAY_MISSING
    if(!base->widget)
#endif
      printf("%s", color);
    base->colormode = colormode;
  }
}
